'''
    Date:   10/26/2024
    Author: Martin E. Liza
    File:   constant_tables.py
    Def:    This file contains spectroscopy data used across the package
'''

# Gladstone-Dale constants 
def karl_2003(): #follows SU2 MutationPP format 
# https://arc.aiaa.org/doi/pdf/10.2514/6.2003-4252
    dict_out = { 'N'  : 0.301E-3,
                 'O'  : 0.182E-3,
                 'NO' : 0.221E-3,
                 'N2' : 0.238E-3,
                 'O2' : 0.190E-3 } #[m3/kg]
    return dict_out

# Polarizability
def polarizability(): #follows SU2 MutationPP format
    dict_out = { #'e-'  : 0.0, 
                 'N+'   : 0.559E-30,
                 'O+'   : 0.345E-30,
                 'NO+'  : 1.021E-30,
                 'N2+'  : 2.386E-30,
                 'O2+'  : 0.238E-30,
                 'N'    : 1.100E-30,
                 'O'    : 0.802E-30,
                 'NO'   : 1.700E-30,
                 'N2'   : 1.7403E-30,
                 'O2'   : 1.5689E-30 } # [m^3]  
    return dict_out 

    
def polarizability_derivatives(molecule='N2'):
    # https://link.springer.com/content/pdf/10.1134/BF03355985.pdf
    """
    Constants below come from Buldakov Paper: 'Temperature Dependence of Polarizability
of Diatomic Homonuclear Molecules'. Reference for each constant are provided
below
    """
    dict_out = { }
    if molecule == 'H2':
        dict_out['zeroth'] = 0.7849E-30 #https://edisciplinas.usp.br/pluginfile.php/4557662/mod_resource/content/1/CRC%20Handbook%20of%20Chemistry%20and%20Physics%2095th%20Edition.pdf
        dict_out['first'] = 0.90E-30 #https://www.tandfonline.com/doi/abs/10.1080/00268978000103191
        dict_out['second'] = 0.49E-30 #https://www.tandfonline.com/doi/abs/10.1080/00268978000103191 
        dict_out['third'] = -0.85E-30 #https://www.tandfonline.com/doi/abs/10.1080/00268978000103191

    if molecule == 'N2':
        dict_out['zeroth'] = 1.7801E-30 #https://edisciplinas.usp.br/pluginfile.php/4557662/mod_resource/content/1/CRC%20Handbook%20of%20Chemistry%20and%20Physics%2095th%20Edition.pdf
        dict_out['first'] = 1.86E-30 #M. A. Buldakov, B. V. Korolev, I. I. Matrosov, and T. N. Popova, Opt. Spektrosk. 63, 775 (1987) MISSING REFERENCE 
        dict_out['second'] = 1.2E-30 #M. A. Buldakov, B. V. Korolev, I. I. Matrosov, and T. N. Popova, Opt. Spektrosk. 63, 775 (1987) MISSING REFERENCE 
        dict_out['third'] = -4.6E-30 #https://pubs.aip.org/aip/jcp/article-abstract/78/9/5287/777120/Theoretical-study-of-the-effects-of-vibrational?redirectedFrom=fulltext

    if molecule == 'O2':
        dict_out['zeroth'] = 1.6180E-30 #https://edisciplinas.usp.br/pluginfile.php/4557662/mod_resource/content/1/CRC%20Handbook%20of%20Chemistry%20and%20Physics%2095th%20Edition.pdf
        dict_out['first'] = 1.76E-30 #M. A. Buldakov, B. V. Korolev, I. I. Matrosov, and T. N. Popova, Opt. Spektrosk. 63, 775 (1987) MISSING REFERENCE 
        dict_out['second'] = 3.4E-30 #M. A. Buldakov, B. V. Korolev, I. I. Matrosov, and T. N. Popova, Opt. Spektrosk. 63, 775 (1987) MISSING REFERENCE 
        dict_out['third'] = -23.7E-30 #https://pubs.aip.org/aip/jcp/article-abstract/100/2/1297/482621/Frequency-dependent-polarizabilities-of-O2-and-van?redirectedFrom=fulltext

    return dict_out #[m^3]

def kerl_interpolation(molecule='N2'):
    #https://onlinelibrary.wiley.com/doi/10.1002/bbpc.19920960517
    # Check reference in paper
    dict_out = { }
    if molecule == 'H2':
        dict_out['groundPolarizability'] = 0.80320E-30 #[m^3]
        dict_out['groundFrequency'] = 2.1399E16 # [1/s]
        dict_out['b'] = 5.87E-6 # [1/K]
        dict_out['c'] = 7.544E-9 # [1/K^2]

    if molecule == 'N2':
        dict_out['groundPolarizability'] = 1.7406E-30 #[m^3]
        dict_out['groundFrequency'] = 2.6049E16 # [1/s]
        dict_out['b'] = 1.8E-6 # [1/K]
        dict_out['c'] = 0.0 
    if molecule == 'O2':
        dict_out['groundPolarizability'] = 1.5658E-30 #[m^3]
        dict_out['groundFrequency'] = 2.1801E16 # [1/s]
        dict_out['b'] = -2.369E-6 # [1/K]
        dict_out['c'] = 8.687E-9# [1/K^2]
    if molecule == 'Air':
        dict_out['groundPolarizability'] = 1.6970E-30 #[m^3]
        dict_out['groundFrequency'] = 2.47044E16 # [1/s]
        dict_out['b'] = 10.6E-6 # [1/K]
        dict_out['c'] = 7.909E-9# [1/K^2]

    return dict_out

def spectroscopy_constants(molecule='N2'):
    # https://doi.org/10.1063/1.2436891
    # minimun electronic energy T_e [cm^-1]
    # harmonic frequency (omega_e)  [cm^-1]
    # first anharmonic correction (omega_e x_e) [cm^-1]
    # (omega_e y_e)
    # equilibrium rotational constant (B_e)
    # anharmonic correction to the rotational constant (alpha_e) [cm^-1]
    # centrifugal distortion constant (D_e) [cm^-1]
    # binding energy (D_o) [eV]
    # equilibrium internuclear distance(R_e) [\r{A}] 
    # ionization potential (IP) [eV]

    dict_out = { }

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=NO%2B&Units=SI&cDI=on
    if molecule == 'NO+':
        dict_out['omega_e'] = 2376.72
        dict_out['omega_xe'] = 16.255
        dict_out['omega_ye'] = -0.01562
        dict_out['B_e'] = 1.997195
        dict_out['alpha_e'] = 0.018790
        dict_out['D_e'] = 6.64E-6
        dict_out['r_e'] = 1.06322E-10 #[m] 

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=N2%2B&Units=SI&cDI=on
    if molecule == 'N2+':
        dict_out['omega_e'] = 2207.0115
        dict_out['omega_xe'] = 16.0616
        dict_out['omega_ye'] = -0.04289
        dict_out['B_e'] = 1.93176
        dict_out['alpha_e'] = 0.0181
        dict_out['D_e'] = 6.10E-6
        dict_out['r_e'] = 1.11642E-10 

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=O2%2B&Units=SI&cDI=on
    if molecule == 'O2+':
        dict_out['omega_e'] = 1905.892
        dict_out['omega_xe'] = 16.489
        dict_out['omega_ye'] = 0.02057
        dict_out['B_e'] = 1.689824
        dict_out['alpha_e'] = 0.019363
        dict_out['D_e'] = 5.32E-6
        dict_out['r_e'] = 1.1164E-10 

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=NO&Units=SI&cDI=on
    if molecule == 'NO':
        dict_out['omega_e'] = 1904.1346
        dict_out['omega_xe'] = 14.08836
        dict_out['omega_ye'] = 0.01005
        dict_out['B_e'] = 1.704885
        dict_out['alpha_e'] = 0.0175416
        dict_out['D_e'] = 0.54E-6 
        dict_out['r_e'] = 1.15077E-10 

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=N2&Units=SI&cDI=on
    if molecule == 'N2':
        dict_out['omega_e'] = 2358.57
        dict_out['omega_xe'] = 14.324
        dict_out['omega_ye'] = -0.00226
        dict_out['B_e'] = 1.998241
        dict_out['alpha_e'] = 0.017318
        dict_out['D_e'] = 5.9E-6
        dict_out['r_e'] = 1.2126E-10 

    #https://webbook.nist.gov/cgi/cbook.cgi?Name=O2&Units=SI&cDI=on
    if molecule == 'O2':
        dict_out['omega_e'] = 1580.161
        dict_out['omega_xe'] = 11.95127
        dict_out['omega_ye'] = 0.0458489
        dict_out['B_e'] = 1.44562
        dict_out['alpha_e'] = 0.0159305
        dict_out['D_e'] = 4.839E-6 
        dict_out['r_e'] = 1.20752E-10 

    return dict_out #wavenumber units  #[cm^-1]



'''
    Date:   08/27/2023
    Author: Martin E. Liza
    File:   aerodynamics_functions.py
    Def:    Contains aerodynamics helper functions. 
'''
import os 
import molmass
import numpy as np 
import scipy.constants as s_consts

# Sutherland law
def sutherland_law(temperature_K): 
    # https://doc.comsol.com/5.5/doc/com.comsol.help.cfd/cfd_ug_fluidflow_high_mach.08.27.html
    viscosity_ref     = 1.716E-5   # [kg/m*s] 
    temperature_ref   = 273.0      # [K]
    sutherland_const  = 111.0      # [K] 
    dynamic_viscosity = ( viscosity_ref * 
                         (temperature_k / temperature_ref) ** (3/2)  *
                         (temperature_ref + sutherland_const) / 
                         (temperature_K + sutherland_const) ) 

    return dynamic_viscosity # [kg/m*s] 


# Air atomic mass
def air_atomic_mass():
    molecules = ['N+', 'O+', 'NO+', 'N2+', 'O2+', 'N', 'O', 'NO', 'N2', 'O2'] 
    air_atomic_dict = { }
    for i in molecules:
        air_atomic_dict[i] = molmass.Formula(i).mass

    return air_atomic_dict #[g/mol]


# Speed of sound 
def speed_of_sound(temperature_K, adiabatic_indx=1.4):
    gas_const          = s_consts.R                         # [J/mol*K]
    air_atomic_mass    = air_atomic_mass()                  # [g/mol]
    air_molecular_mass = (0.7803 * air_atomic_mass['N2'] +  # [kg/mol]
                          0.2099 * air_atomic_mass['O2'] + 
                          0.0003 * air_atomic_mass['CO2']) * 1E-3 
    spd_of_sound       = np.sqrt( adiabatic_indx * temperature_K * 
                                  gas_const / air_molecular_mass )  
    return spd_of_sound # [m/s]


# Normal shock relations  
def normal_shock_relations(mach_1, adiabatic_indx=1.4):
    # REF: https://www.grc.nasa.gov/www/k-12/airplane/normal.html
    # NOTE: var_r = var_1 / var_2 = var_preshock / var_postshock = [ ] 
    gamma_minus   = adiabatic_indx - 1
    gamma_plus    = adiabatic_indx + 1
    mach_11       = mach_1 ** 2
    mach_2        = np.sqrt( (gamma_minus * mach_11 + 2) /
                             (2 * adiabatic_indx * mach_11 - gamma_minus) )
    pressure_r    = (2 * adiabatic_indx * mach_11 - gamma_minus) / gamma_plus
    temperature_r = ( (2 * adiabatic_indx * mach_11 - gamma_minus) *
                      (gamma_minus * mach_11 + 2) / (gamma_plus**2 * mach_11) )
    density_r     = gamma_plus * mach_11 / (gamma_minus * mach_11 + 2)

    # Return Dictionary 
    normal_shock_dict = { 'mach_2'        : mach_2,
                          'pressure_r'    : pressure_r, 
                          'temperature_r' : temperature_r,
                          'density_r'     : density_r }
    return normal_shock_dict # [ ]

# Oblique shock relations
def oblique_shock_relations(mach_1, shock_angle_deg, adiabatic_indx=1.4):
# REF : Modern Compressible Flows With Historical Ref., eq 4.7 - 4.11 
# NOTE: Equations only work for weak shocks 
# Note ratio = var_1 / var_2
    shock_angle = np.radians(shock_angle_deg)  # radians
    sin2_shock  = np.sin(shock_angle)**2
    mach_n1     = mach_1 * np.sin(shock_angle) # normal mach number 
    mach_n11    = mach_n1 ** 2  # normal mach number square 
    # Calculates Deflection angle (Eq. 4.17) 
    tan_deflection_ang   = ( (2 / np.tan(shock_angle)) *  ( (mach_n11 - 1) / 
                           (mach_1**2 * (adiabatic_indx + 
                           np.cos(2 * shock_angle)) + 2) ) ) 
    deflection_angle_deg = np.degrees(np.arctan(1 / tan_deflection_ang)) 
    # Calculates properties downstream the shock  
    density_r     = ( ((adiabatic_indx + 1) * mach_n1**2) / 
                     ((adiabatic_indx - 1) * mach_n1**2 + 2) )
    pressure_r    = (1 + 2 * adiabatic_indx * (mach_n1**2 - 1) / 
                     (adiabatic_indx + 1) )
    temperature_r = pressure_r * 1 / density_r 
    # Calculates mach 2 
    mach_n2 = np.sqrt( (mach_n1**2 + (2 / (adiabatic_indx - 1))) / 
            ((2 * adiabatic_indx / (adiabatic_indx - 1)) * mach_n1**2 - 1) )
    mach_2  = mach_n2 / np.sin(np.radians(shock_angle_deg - 
                              deflection_angle_deg)) 
    # Dictionary 
    oblique_shock_dict = { 'mach_2'                : mach_2,
                           'pressure_r'            : pressure_r, 
                           'temperature_r'         : temperature_r,
                           'density_r'             : density_r, 
                           'deflection_angle_degs' : deflection_angle_deg } 
    return oblique_shock_dict





#!/opt/homebrew/bin/python3.9
'''
    Date:   03/26/2023
    Author: Martin E. Liza
    File:   aero_optics.py
    Def:    Contains aero optics functions. 
'''
from ambiance import Atmosphere
import numpy as np
import scipy.constants as s_consts 

# My Packages 
from haot.src import aerodynamic_functions as aero
from haot.src import constants_tables 
from haot.src import quantum_mechanics as quantum

def gas_density(density_dict): # density_dict [kg/m^3]
    gas_amu_weight  = aero.air_atomic_mass()  # [g/mol]  
    avogadro_number = s_consts.N_A               # [particles/mol]  
    gas_density     = { }
    
    for i in density_dict: 
        gas_density[i] = (density_dict[i] * 10**3 * 
                               avogadro_number /
                               gas_amu_weight[i]) # [particles/m^3] 

    return gas_density #[particles/m^3] 

def index_of_refraction(gas_density_dict):
    pol_consts         = constants_tables.polarizability() # [m^3] 
    dielectric_const_0 = s_consts.epsilon_0                # [F/m] 
    density            = gas_density(gas_density_dict)     # [particles/m3]    
    n_const            = { }                               # [ ] 
    # Convert cgs to SI
    alpha_si = lambda x : x * (4 * np.pi * dielectric_const_0) #[F m2]

    for i in gas_density_dict: 
        # Convert alpha_cgs to alpha_si 
        alpha      = alpha_si(pol_consts[i]) 
        n_const[i] = (alpha * density[i]) #(a_i N_i)

    # add all n_i 
    temp = 0.0
    for i in n_const.values():
        temp += i 

    n_return = { }
    n_return['dilute'] = 1 + temp / (2 * dielectric_const_0) 
    n_temp             = temp / (3 * dielectric_const_0)
    n_return['dense']  = ( (2 * n_temp + 1) / (1 - n_temp) )**0.5
    # Note np.sqrt(-1) does not detect complex 

    return n_return 

def dielectric_material_const(n_const): 
    # n ~ sqrt(e_r)  
    dielectric_const_0   = s_consts.epsilon_0 # [F/m] 
    dielectric           = { }
    dielectric['dilute'] = dielectric_const_0 * n_const['dilute']**2
    dielectric['dense']  = dielectric_const_0 * n_const['dense']**2
    return dielectric 

def optical_path_length(n_solution, distance):
    OPL           = { }
    OPL['dilute'] = n_solution['dilute'] * distance
    OPL['dense']  = n_solution['dense'] * distance
    return OPL 

def tropina_aproximation(vibrational_number, rotational_number, molecule):
    electron_mass = s_consts.m_e
    electron_charge = s_consts.e
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    #resonance_distance = omega_gi - omega


# Calculate polarizability (uses equation 4 from the paper)
def buldakov_expansion(vibrational_number, rotational_number, molecule):
    # Load constants
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    derivative_const = constants_tables.polarizability_derivatives(molecule)
    be_we = spectroscopy_const['B_e'] / spectroscopy_const['omega_e']

    # Dunham potential energy constants
    (a_0, a_1, a_2) = quantum.potential_dunham_coef_012(molecule)
    a_3 = quantum.potential_dunham_coeff_m(a_1, a_2, 3)

    rotational_degeneracy = rotational_number * (rotational_number + 1)
    vibrational_degeneracy = 2 * vibrational_number + 1

    # Split in terms
    tmp_1 = be_we
    tmp_1 *= (-3 * a_1 * derivative_const['first'] +
             derivative_const['second'])
    tmp_1 *= vibrational_degeneracy 
    tmp_1 *= 1/2

    tmp_2 = be_we**2
    tmp_2 *= derivative_const['first']
    tmp_2 *= rotational_degeneracy
    tmp_2 *= 4

    tmp_31a = 7
    tmp_31a += (15 * vibrational_degeneracy**2) 
    tmp_31a *= a_1**3
    tmp_31a *= -3/8

    tmp_31b = 23
    tmp_31b += (39 * vibrational_degeneracy**2) 
    tmp_31b *= a_2
    tmp_31b *= a_1
    tmp_31b *=  1/4

    tmp_31c = 5
    tmp_31c += vibrational_degeneracy**2 
    tmp_31c *= a_3
    tmp_31c *= -15/4

    tmp_31 = derivative_const['first'] * (tmp_31a + tmp_31b + tmp_31c)

    tmp_32a = 7
    tmp_32a += (15 * vibrational_degeneracy**2)
    tmp_32a *= a_1**2
    tmp_32a *= 1/8

    tmp_32b = 5
    tmp_32b += vibrational_degeneracy**2
    tmp_32b *= a_2
    tmp_32b *- -3/4

    tmp_32 = derivative_const['second'] * (tmp_32a + tmp_32b)

    tmp_33 = 7
    tmp_33 += (15 * vibrational_degeneracy**2)
    tmp_33 *= a_1
    tmp_33 *= derivative_const['third']
    tmp_33 *= -1/24

    tmp_3 = (tmp_31 + tmp_32 + tmp_33) * be_we**2

    tmp_41 = 1 - a_2
    tmp_41 *= 24
    tmp_41 += (27 * a_1 * (1 + a_1))
    tmp_41 *= derivative_const['first']

    tmp_42 = (1 + 3 * a_1)
    tmp_42 *= derivative_const['second']
    tmp_42 *= -3

    tmp_43 = 1/8 * derivative_const['third']

    tmp_4 = (tmp_41 + tmp_42 + tmp_43) 
    tmp_4 *= rotational_degeneracy
    tmp_4 *= vibrational_degeneracy
    tmp_4 *= be_we**3

    return derivative_const['zeroth'] + tmp_1 + tmp_2 + tmp_3 + tmp_4


# Calculate polarizability as temperature
"""
    DOI: 10.1002/bbpc.19920960517 
    DOI: 10.1134/BF03355985
"""
def kerl_polarizability_temperature(*args, **kargs):
    if args:
        temperature_K = args[0] 
        molecule = args[1]
        wavelength_nm = args[2]

    if kargs:
        temperature_K = kargs['temperature_K']
        molecule = kargs['molecule']
        wavelength_nm = kargs['wavelength_nm']

    # Check sizes
    mean_const = constants_tables.kerl_interpolation(molecule)
    angular_frequency = (2 * np.pi * s_consts.speed_of_light /
                         (wavelength_nm * 1E-9))

    tmp = mean_const['c'] * temperature_K**2
    tmp += mean_const['b'] * temperature_K
    tmp += 1
    tmp *= mean_const['groundPolarizability']
    tmp /= (1 - (angular_frequency / mean_const['groundFrequency'])**2)

    return tmp

# http://walter.bislins.ch/bloge/index.asp?page=Deriving+Equations+for+Atmospheric+Refraction
def atmospheric_index_of_refraction(altitude, vaporPressure=0): 
    atmospheric_prop = Atmosphere(altitude)
    temperature      = atmospheric_prop.temperature #[K]
    pressure         = atmospheric_prop.pressure * 0.01 #[mbar]
    K_1              = 79 #[K/mbar]
    K_2              = 4800 #[K]

    refractivity =  K_2 * vaporPressure / temperature
    refractivity += pressure
    refractivity *= (K_1 / temperature)
    refractivity *= 10**-6

    return refractivity + 1 

def atmospheric_gladstoneDaleConstant(altitude=0.0, gas_composition_dict=None):
    atmospheric_prop = Atmosphere(altitude)
    density          = atmospheric_prop.density * 1E3   #[g/m3]
    num_density      = atmospheric_prop.number_density  #[particles/m3]
    gladstone_const  = Gladstone_Dale()                 #[m3/kg]
    avogadro_number  = s_consts.N_A                     #[particles/mol]

    if gas_composition_dict == None:
        gas_composition_dict = { }
        gas_composition_dict['N'] = 0.0
        gas_composition_dict['O'] = 0.0
        gas_composition_dict['NO'] = 0.0
        gas_composition_dict['N2'] = 0.79
        gas_composition_dict['O2'] = 0.21

    tmp = 0
    for i in gas_composition_dict.keys():
        tmp += gas_composition_dict[i] * gladstone_const[i] 

    return tmp 

def Gladstone_Dale(gas_density_dict=None): # [kg/m3
    gas_amu_weight   = aero.air_atomic_mass()       # [g/mol]
    avogadro_number  = s_consts.N_A                 # [particles/mol]
    dielectric_const = s_consts.epsilon_0           # [F/m]
    gd_consts        = constants_tables.karl_2003() # [m3/kg]
    pol_consts       = constants_tables.polarizability() #[m^3]

    # Convert CGS to SI 
    pol_consts.update({n: 4 * np.pi * dielectric_const * pol_consts[n]
                       for n in pol_consts.keys()}) # [Fm^2]

    # Calculate Gladstone dale
    gladstone_dale_const = { }
    for i in pol_consts:
        gladstone_dale_const[i] = ( pol_consts[i] / (2 * dielectric_const) * 
                (avogadro_number / gas_amu_weight[i]) * 1E3 ) #[m3/kg]

    gladstone_dale_dict = { }
    if not gas_density_dict:
        return gladstone_dale_const #[m^3/kg]
    else:
        gladstone_dale_dict['gladstone_dale'] = 0.0
        for i in gas_density_dict:
            gladstone_dale_dict[i] = ((gladstone_dale_const[i] *
                                       gas_density_dict[i]) /
                                      sum(gas_density_dict.values()))
            gladstone_dale_dict['gladstone_dale'] += gladstone_dale_const[i] * gas_density_dict[i]
        gladstone_dale_dict['gladstone_dale'] /= sum(gas_density_dict.values())

        return gladstone_dale_dict #[m3/kg]



'''
    Date:   11/26/2024
    Author: Martin E. Liza
    File:   quantum_mechanics.py
    Def:    Contains Quantum Mechanics functions.
'''
import molmass
import numpy as np
import scipy.constants as s_consts 
from haot.src import aerodynamic_functions as aero
from haot.src import constants_tables 

# Unit Conversions
def wavenumber_to_electronvolt(wavenumber_cm):
    """Convert wavenumber [cm^-1] to energy in Joules [J]."""
    return wavenumber_to_joules(wavenumber_cm) / s_consts.eV #[eV]

def wavenumber_to_joules(wavenumber_cm):
    """Convert wavenumber [cm^-1] to energy in electron volts [eV]."""
    return wavenumber_cm * s_consts.c * 100 * s_consts.h #[J]

def molarmass_to_kilogram(molarmass_gmol):
    """Convert molar mass [g/mol] to [kg]."""
    return molarmass_gmol * 1E-3 / s_consts.N_A

def zero_point_energy(molecule):
    """Calculate zero-point energy based on spectroscopy constants. 
    (Ref: Irikura https://doi.org/10.1063/1.2436891)"""
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    scope_var = (spectroscopy_const['alpha_e'] *
                 spectroscopy_const['omega_e'] /
                 spectroscopy_const['B_e'])
    zpe = spectroscopy_const['omega_e'] / 2
    zpe -= spectroscopy_const['omega_xe'] / 2
    zpe += spectroscopy_const['omega_ye'] / 8
    zpe += spectroscopy_const['B_e'] / 4
    zpe += scope_var / 12 
    zpe += scope_var**2 / (144 * spectroscopy_const['B_e'])
    return zpe #[1/cm] 

def vibrational_partition_function(vibrational_number, temperature_K, molecule):
    """Calculates the vibrational partition function, (harmonic terms only)"""
    z_vib = 0.0
    for v in range(vibrational_number + 1):
        z_vib += boltzman_factor(temperature_K, molecule, vibrational_number=v)
    return z_vib

def rotational_partition_function(rotational_number, temperature_K, molecule):
    """Calculates the rotational partition function, (harmonic terms only)"""
    z_rot = 0.0
    for j in range(rotational_number + 1):
        z_rot += boltzman_factor(temperature_K, molecule, rotational_number=j)
    return z_rot

def born_oppenheimer_partition_function(vibrational_number,
                                        rotational_number,
                                        temperature_K,
                                        molecule):
    """Calculates the partition function using the Born-Oppenheimer
    approximation"""
    z_bo = 0.0
    for j in range(rotational_number + 1):
        for v in range(vibrational_number + 1):
            z_bo += boltzman_factor(temperature_K, molecule,
                                    vibrational_number=v,
                                    rotational_number=j,
                                    born_opp_flag=True)
    return z_bo


def potential_dunham_coef_012(molecule):
    """Calculates the 0th, 1st, and 2nd Dunham potential coefficients.
    Using: Ogilvie (https://doi.org/10.1016/0022-2852(76)90323-4)
    and Herschbach (https://doi.org/10.1063/1.1731952)."""
    spectroscopy_const = constants_tables.spectroscopy_constants(molecule)
    a_0 = (spectroscopy_const['omega_e']**2 /
           (4 * spectroscopy_const['B_e']))
    a_1 = -(spectroscopy_const['alpha_e'] * spectroscopy_const['omega_e'] /
               (6 * spectroscopy_const['B_e']**2) + 1)
    a_2 = ((5/4) * a_1**2 - (2/3) *
           (spectroscopy_const['omega_xe'] / spectroscopy_const['B_e'])) 
    return (a_0, a_1, a_2)

def potential_dunham_coeff_m(a_1, a_2, m):
    """Calculates the higher order Dunham potential coefficients, using
    Morizadeh work (https://doi.org/10.1016/j.theochem.2003.12.003)."""
    tmp = (12 / a_1)**(m - 2)
    tmp *= (2**(m + 1) - 1)
    tmp *= (a_2 / 7)**(m - 1)
    for i in range(m - 2):
        tmp *= (1 / (m + 2 - i))

    return tmp


def boltzman_factor(temperature_K, molecule, vibrational_number=None,
                    rotational_number=None, born_opp_flag=False):
    """Calculates the Boltzman factor at a given vibrational_number and/or
    rotational_number. If the born_opp_flag is provided, it will calculate the
    total energy using the Born-Oppenheimer approximation"""
    # Initialize energy terms, degeneracy and thermal beta
    energy_vib_k = 0
    energy_rot_k = 0
    degeneracy_rotation = 1
    thermal_beta = 1 / (s_consts.k * temperature_K)

    # Calculates Energy levels
    if not born_opp_flag:
        if vibrational_number is not None:
            energy_vib_k = vibrational_energy_k(vibrational_number, molecule)
        if rotational_number is not None:
            energy_rot_k = rotational_energy_k(rotational_number, molecule)
            degeneracy_rotation = 2 * rotational_number + 1
        tot_energy = wavenumber_to_joules(energy_vib_k + energy_rot_k)
    else:
        degeneracy_rotation = 2 * rotational_number + 1
        tot_energy = wavenumber_to_joules(born_oppenheimer_approximation(vibrational_number,
                                                        rotational_number,molecule))
    return degeneracy_rotation * np.exp(-tot_energy * thermal_beta)


def distribution_function(temperature_K, molecule, vibrational_number=None,
                    rotational_number=None, born_opp_flag=False):
    """Compute the population distribution function."""
    # Calculates partition functions if vibrational or rotational numbers are provided
    if not born_opp_flag:
        z_rot = 1
        z_vib = 1
        if vibrational_number is not None:
            z_vib = vibrational_partition_function(vibrational_number, temperature_K,
                                          molecule)
        if rotational_number is not None:
            z_rot = rotational_partition_function(rotational_number, temperature_K,
                                         molecule)
        z_tot = z_rot * z_vib
    else:
        z_tot = born_oppenheimer_partition_function(vibrational_number,
                                       rotational_number, temperature_K,
                                       molecule)

    # Create the distribution array based on the inputs provided
    if vibrational_number and rotational_number:
        tmp = np.zeros([vibrational_number + 1, rotational_number + 1])
        for j in range(rotational_number + 1):
            for v in range(vibrational_number + 1):
                tmp[v][j] = boltzman_factor(temperature_K=temperature_K,
                                            molecule=molecule,
                                            vibrational_number=v,
                                            rotational_number=j,
                                            born_opp_flag=born_opp_flag)
    elif vibrational_number:
        tmp = np.zeros(vibrational_number + 1)
        for v in range(vibrational_number + 1):
            tmp[v] = boltzman_factor(temperature_K=temperature_K,
                                        molecule=molecule,
                                        vibrational_number=v)

    elif rotational_number:
        tmp = np.zeros(rotational_number + 1)
        for j in range(rotational_number + 1):
            tmp[j] = boltzman_factor(temperature_K=temperature_K,
                                        molecule=molecule,
                                        rotational_number=j)
    return tmp / z_tot

def born_oppenheimer_approximation(vibrational_number, rotational_number,
                                   molecule):
    """Calculates the energy at a rotational and vibrational quantum number,
    using the Born-Oppenheimer approximation."""
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)

    vib_levels = vibrational_number + 1/2
    rot_levels = rotational_number * (rotational_number + 1)

    # Harmonic vibration and rotation terms
    harmonic = spectroscopy_constants['omega_e'] * vib_levels
    harmonic += spectroscopy_constants['B_e'] * rot_levels

    # Anharmonic vibration and rotation terms 
    anharmonic = spectroscopy_constants['omega_xe'] * vib_levels**2
    anharmonic += spectroscopy_constants['D_e'] * rot_levels**2

    # Interaction between vibration and rotation modes
    interaction = spectroscopy_constants['alpha_e'] * vib_levels * rot_levels

    return harmonic - anharmonic - interaction #[cm^1]

def vibrational_energy_k(vibrational_number, molecule):
    """Calculates the vibrational energy at a given vibrational quantum number,
    using for the harmonic terms"""
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)
    # Calculates the vibrational energy in units of wave number
    vib_levels = vibrational_number + 1/2
    return spectroscopy_constants['omega_e'] * vib_levels #[cm^-1]

def rotational_energy_k(rotational_number, molecule):
    """Calculates the rotational energy at a given rotational quantum number,
    using for the harmonic terms"""
    spectroscopy_constants = constants_tables.spectroscopy_constants(molecule)
    # Calculates the rotational energy in units of wave number
    rot_levels = rotational_number * (rotational_number + 1)
    return spectroscopy_constants['B_e'] * rot_levels #[cm^-1]

def reduced_mass_kg(molecule_1, molecule_2):
    """Calculates the molar reduced mass and returns it in kg of two
    elements"""
    m_1 = molmass.Formula(molecule_1).mass
    m_2 = molmass.Formula(molecule_2).mass
    mu = m_1 * m_2 / (m_1 + m_2)

    return molarmass_to_kilogram(mu)

# TODO: Missing Translational Energy
def tranlational_energy(principal_number_x, principal_number_y,
                        principal_number_z):
    print("TODO: Missing implementation of this function")



from setuptools import setup, find_packages

setup(
    name="haot",
    version="1.0.0-b4",
    author="Martin E. Liza",
    author_email="mliza1191@gmail.com",
    description="Hypersonic Aerodynamic Optics Tools",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/mliza/HAOT",
    packages=find_packages(),
    license="MIT",
    python_requires=">=3.7",
    install_requires=[
        "numpy>=1.20.0",
        "scipy>=1.6.0",
        "molmass>=2024.10.25",
        "ambiance>=1.3.1"
    ],
    entry_points={
        "console_scripts": [
            "hoat-cli=hoat.__main__:main",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)
